# -*- coding: utf-8 -*-
# @Time    : 03/03/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

from sympy import Matrix,  symbols,  diff
from sympy import sin,  cos,  atan,  lambdify,  atan2,  I
import math
import numpy as np
from numpy import real as re
from numpy import imag as im

'Se definen las variables simbolicas para posteriormente utilizarlas en las ecuaciones de cálculo'
'de flujo de P y Q por las lineas'
V_i = symbols('V_i', real=True)
theta_i = symbols('theta_i', real=True)
V_j = symbols('V_j', real=True)
theta_j = symbols('theta_j', real=True)
G = symbols('G', real=True)
B = symbols('B', real=True)

'Real an reactive power inyection'
def Pi():
    """
    Symbolic equation of Pi, in matrix format
    :return: P_i
    """
    #P_i = V_i * ((V_j * (G * cos(theta_i - theta_j) + B * sin(theta_i - theta_j)))) - V_i * ((V_i * (G * cos(theta_i - theta_i) + B * sin(theta_i - theta_i))))
    #P_i = 3 * P_i
    P_i = V_i * V_j * (G * cos(theta_i - theta_j) + B * sin(theta_i - theta_j))
    P_i = 3 * P_i

    return P_i

def Qi():
    """
    Symbolic equation of Qi, in matrix format

    :return: Q_i
    """
    #Q_i = V_i * ((V_j * (G * sin(theta_i - theta_j) - B * cos(theta_i - theta_j)))) - V_i * ((V_i * (G * sin(theta_i - theta_i) - B * cos(theta_i - theta_i))))
    Q_i = V_i * V_j * (G * sin(theta_i - theta_j) - B * cos(theta_i - theta_j))
    Q_i = 3 * Q_i

    return Q_i

'Real an reactive power flows'
def Pij():
    """
    Symbolic equation of Pij, in matrix format

    :return: P_ij
    """
    P_ij = V_i * ((V_j * (G * cos(theta_i - theta_j) + B * sin(theta_i - theta_j)))) - \
           V_i * ((V_i * (G * cos(theta_i - theta_i) + B * sin(theta_i - theta_i))))

    P_ij = 3 * P_ij

    return P_ij

def Qij():
    """
    Symbolic equation of Qij, in matrix format

    :return: Q_ij
    """
    '''
    Vi = np.array(V_i * (cos(theta_i) + I * sin(theta_i)))
    Vj = np.array(V_j * (cos(theta_j) + I * sin(theta_j)))
    Vt = np.array(Vi - Vj)
    Y = np.array(G + I * B)
    I_ij = np.dot(Y, Vt)
    Q_ij = 3 * Vi * np.conj(I_ij)
    '''
    Q_ij = V_i * ((V_j * (G * sin(theta_i - theta_j) - B * cos(theta_i - theta_j)))) - \
           V_i * ((V_i * (G * sin(theta_i - theta_i) - B * cos(theta_i - theta_i))))
    Q_ij = 3 * Q_ij

    return Q_ij

'Circuit current, module and angle'
def Iij():
    """
    Symbolic equation of |Iij|, in matrix format

    :return: Iij
    """
    P_ij = V_i * ((V_j * (G * cos(theta_i - theta_j) + B * sin(theta_i - theta_j)))) - \
           V_i * ((V_i * (G * cos(theta_i - theta_i) + B * sin(theta_i - theta_i))))
    #P_ij = 3 * P_ij
    Q_ij = V_i * ((V_j * (G * sin(theta_i - theta_j) - B * cos(theta_i - theta_j)))) - \
           V_i * ((V_i * (G * sin(theta_i - theta_i) - B * cos(theta_i - theta_i))))

    #Q_ij = 3 * Q_ij

    I_ij = pow(((P_ij) ** 2 + (Q_ij) ** 2),  0.5) / V_i

    return I_ij

def angIij():
    """
    Symbolic equation of Iang_δij, in matrix format

    :return: angI_ij
    """

    P_ij = V_i * ((V_j * (G * cos(theta_i - theta_j) + B * sin(theta_i - theta_j)))) - \
           V_i * ((V_i * (G * cos(theta_i - theta_i) + B * sin(theta_i - theta_i))))

    Q_ij = V_i * ((V_j * (G * sin(theta_i - theta_j) - B * cos(theta_i - theta_j)))) - \
           V_i * ((V_i * (G * sin(theta_i - theta_i) - B * cos(theta_i - theta_i))))

    '''
    angI_ij = theta_i - atan(Q_ij/P_ij)
    '''
    angI_ij = theta_i -atan2(Q_ij,  P_ij)

    return angI_ij

def Iijrec():
    """
    Symbolic equation of Iij_rectangular, in matrix format

    :return: I_ij
    """
    Vi = np.array(V_i * (cos(theta_i) + I * sin(theta_i)))
    Vj = np.array(V_j * (cos(theta_j) + I * sin(theta_j)))

    Vt = np.array(Vi - Vj)

    Y = np.array(G + I * B)

    #I_ij = np.dot(Y, Vt)
    I_ij = np.dot(Y, Vt)

    return I_ij


'----------------------------------------------------------------------------------------------------------------------'
'''
Symbolic equations that form the Jacobian matrix

___________|_____θ_____|_____V_____|
____|Vi|___|____H11____|____H12____|
_____Pi____|____H21____|____H22____|
_____Qi____|____H31____|____H32____|
____Pij____|____H41____|____H42____|
____Qij____|____H51____|____H52____|
___|Iij|___|____H61____|____H62____|
____|Vi|___|____H71____|____H72____|
_____θi____|____H81____|____H82____|
___|Iij|___|____H91____|____H92____|
____δij____|____H01____|____H02____|
'''
'----------------------------------------------------------------------------------------------------------------------'


def H21i_a():
    """
    Symbolic equation of ∂Pi/∂θi part (a), in matrix format
    Eq. available in ''Formulation of Three-Phase State Estimation Problem Using a Virtual Reference'

    :return: H21i_a
    """
    '''
    H21i= Matrix([[diff(P_i()[0], theta_i1), diff(P_i()[0], theta_i2), diff(P_i()[0], theta_i3)],
                  [diff(P_i()[1], theta_i1), diff(P_i()[1], theta_i2), diff(P_i()[1], theta_i3)], 
                  [diff(P_i()[2], theta_i1), diff(P_i()[2], theta_i2), diff(P_i()[2], theta_i3)]])
    '''
    # Eq. (16)
    H21i_a = diff(Pi(), theta_i)

    return H21i_a


def H21i_b():
    """
    Symbolic equation of ∂Pi/∂θi part (b), in matrix format
    Eq. available in ''Formulation of Three-Phase State Estimation Problem Using a Virtual Reference'

    :return: H21i_b
    """
    H21i_b = -Qi()

    return H21i_b

def H21j():
    """
    Symbolic equation of ∂Pi/∂θj, in matrix format

    :return: H21j
    """
    #Eq. (17)
    H21j= diff(Pi(), theta_j)
    return H21j

def H22i_a():
    """
    Symbolic equation of ∂Pi/∂Vi part (a), in matrix format

    :return: H22i_a
    """

    H22i_a= diff(Pi(), V_i)

    return H22i_a


def H22i_b():
    """
    Symbolic equation of ∂Pi/∂Vi part (b), in matrix format

    :return: H22i_b
    """

    H22i_b = Pi()

    return H22i_b

def H22j():
    """
    Symbolic equation of ∂Pi/∂Vj, in matrix format

    :return: H22j
    """
    H22j= diff(Pi(), V_j)

    return H22j

'H31i(∂Qi/∂θi), H31j(∂Qi/∂θj), H32i(∂Qi/∂Vi) and H32j(∂Qi/∂Vj) are the derivatives of reactive injection power at node "i" and "j" respectively'
def H31i_a():
    """
    Symbolic equation of ∂Qi/∂θi part (a), in matrix format
    Eq. available in ''Formulation of Three-Phase State Estimation Problem Using a Virtual Reference

    :return: H31i_a
    """

    #Eq. (19)
    H31i_a= diff(Qi(), theta_i)

    return H31i_a

def H31i_b():
    """
    Symbolic equation of ∂Qi/∂θi part (b), in matrix format

    :return: H31i_b
    """
    H31i_b = Pi()

    return H31i_b

def H31j():
    """
    Symbolic equation of ∂Qi/∂θj, in matrix format

    :return: H31j
    """
    #Eq. (20)
    H31j= diff(Qi(), theta_j)
    return H31j

def H32i_a():
    """
    Symbolic equation of ∂Qi/∂Vi part (a), in matrix format

    :return: H32i_a
    """

    H32i_a = diff(Qi(), V_i)

    return H32i_a

def H32i_b():
    """
    Symbolic equation of ∂Qi/∂Vi part (b), in matrix format

    :return: H32i_b
    """

    H32i_b= Qi()

    return H32i_b

def H32j():
    """
    Symbolic equation of ∂Qi/∂Vj, in matrix format

    :return: H32j
    """
    H32j = diff(Qi(), V_j)

    return H32j

'----------------------------------------------------------------------------------------------------------------------'
'H41i(∂Pij/∂θi), H41j(∂Pij/∂θj), H42i(∂Pij/∂Vi) and H42j(∂Pij/∂Vj) are the derivatives of the active power of flow at node "i" and "j" respectively'
def H41i():
    """
    Symbolic equation of ∂Pij/∂θi, in matrix format

    :return: H41i
    """
    H41i = diff(Pij(), theta_i)

    return H41i

def H41j():
    """
    Symbolic equation of ∂Pij/∂θj, in matrix format

    :return: H41j
    """

    H41j = diff(Pij(), theta_j)

    return H41j

def H42i():
    """
    Symbolic equation of ∂Pij/∂Vi, in matrix format

    :return: H42i
    """

    H42i= diff(Pij(), V_i)

    return H42i

def H42j():
    """
    Symbolic equation of ∂Pij/∂Vj, in matrix format

    :return: H42j
    """
    H42j= diff(Pij(), V_j)

    return H42j

'----------------------------------------------------------------------------------------------------------------------'
'H51i(∂Qij/∂θi), H51j(∂Qij/∂θj), H52i(∂Qij/∂Vi) and H52j(∂Qij/∂Vj)'
'Are the derivatives of the reactive power of flow at node "i" and "j" respectively'
def H51i():
    """
    Symbolic equation of ∂Qij/∂θi, in matrix format

    :return: H51i
    """
    H51i = diff(Qij(), theta_i)

    return H51i

def H51j():
    """
    Symbolic equation of ∂Qij/∂θj, in matrix format

    :return: H51j
    """
    H51j = diff(Qij(), theta_j)

    return H51j

def H52i():
    """
    Symbolic equation of ∂Qij/∂Vi, in matrix format

    :return: H52i
    """

    H52i = diff(Qij(), V_i)
    return H52i

def H52j():
    """
    Symbolic equation of ∂Qij/∂Vj, in matrix format

    :return: H52j
    """

    H52j = diff(Qij(), V_j)

    return H52j

'----------------------------------------------------------------------------------------------------------------------'
'H61i(∂|Iij|/∂θi), H61j(∂|Iij|/∂θj), H62i(∂|Iij|/∂Vi) and H62j(∂|Iij|/∂Vj)'
'Are the derivatives of the current module between node "i" and "j" respectively'
def H61i():
    """
    Symbolic equation of ∂|Iij|/∂θi, in matrix format

    :return: H61i
    """
    H61i = diff(Iij(), theta_i)

    return H61i

def H61j():
    """
    Symbolic equation of ∂|Iij|/∂θj, in matrix format

    :return: H61j
    """
    H61j= diff(Iij(), theta_j)

    return H61j

def H62i():
    """
    Symbolic equation of ∂|Iij|/∂Vi, in matrix format

    :return: H62i
    """

    H62i= diff(Iij(), V_i)

    return H62i

def H62j():
    """
    Symbolic equation of ∂|Iij|/∂Vj, in matrix format

    :return: H62j
    """

    H62j= diff(Iij(), V_j)

    return H62j


def H61i_rec():
    """
    Symbolic equation of ∂Iijrec/∂θi, in matrix format

    :return: H62j
    """

    H61i= diff(Iijrec(), theta_i)

    return H61i

def H61j_rec():
    """
    Symbolic equation of ∂Iijrec/∂θj, in matrix format

    :return: H62j
    """

    H61j= diff(Iijrec(), theta_j)

    return H61j

def H62i_rec():
    """
    Symbolic equation of ∂Iijrec/∂θi, in matrix format

    :return: H62j
    """

    H62i= diff(Iijrec(), V_i)

    return H62i

def H62j_rec():
    """
    Symbolic equation of ∂Iijrec/∂θj, in matrix format

    :return: H62j
    """

    H62j= diff(Iijrec(), V_j)

    return H62j

'----------------------------------------------------------------------------------------------------------------------'
'H01i(∂δij/∂θi), H01j(∂δij/∂θj) , H02i(∂δij/∂Vi) and H02j(∂δij/∂Vj)'
'Are the derivatives of the current angle between node "i" and "j" respectively'
def H01i():
    """
    Symbolic equation of ∂δij/∂θi, in matrix format

    :return: H01i
    """
    H01i= diff(angIij(), theta_i)

    return H01i

def H01j():
    """
    Symbolic equation of ∂δij/∂θj, in matrix format

    :return: H01j
    """
    H01j= diff(angIij(), theta_j)

    return H01j

def H02i():
    """
    Symbolic equation of ∂δij/∂Vi, in matrix format

    :return: H02i
    """
    H02i = diff(angIij(), V_i)

    return H02i

def H02j():
    """
    Symbolic equation of ∂δij/∂Vj, in matrix format

    :return: H02j
    """
    H02j = diff(angIij(), V_j)
    return H02j
def sym_func_Pi_Qi():
    """
    Symbolic equations of active and reactive power at node i
    P_i = lambdify(Y_Vij, P_i(), 'numpy')
    Q_i = lambdify(Y_Vij, Q_i(), 'numpy')

    :return:P_i, Q_i
    """
    Y_Vij = [G, B,
             V_i, theta_i,
             V_j, theta_j]
    # (Pi)
    P_i = lambdify(Y_Vij, Pi(), 'numpy')
    # (Qi)
    Q_i = lambdify(Y_Vij, Qi(), 'numpy')

    return P_i, Q_i

def sym_func_Pij_Qij():
    """
    Symbolic equations of active and reactive power between node i and j
    P_ij = lambdify(Y_Vij, P_ij(), 'numpy')
    Q_ij = lambdify(Y_Vij, Q_ij(), 'numpy')

    :return: P_ij, Q_ij
    """
    Y_Vij = [G,B,
             V_i,theta_i,
             V_j,theta_j]
    # (Pij)
    P_ij = lambdify(Y_Vij, Pij(), 'numpy')
    # (Qij)
    Q_ij = lambdify(Y_Vij, Qij(), 'numpy')

    return P_ij, Q_ij

def sym_func_Iij():
    """
    Symbolic equations of current module and angle between node i and j
    I_ij = lambdify(Y_Vij,  Iij(),  'numpy')
    angI_ij = lambdify(Y_Vij,  angIij(),  'numpy')

    :return:I_ij, angI_ij
    """
    Y_Vij = [G,  B,
             V_i,  theta_i,
             V_j,  theta_j]
    # (Iij)
    I_ij = lambdify(Y_Vij, Iij(), 'numpy')
    # (δij-PMU)
    angI_ij = lambdify(Y_Vij, angIij(), 'numpy')

    return I_ij, angI_ij

def sym_func_Iij_rec():
    """
    Symbolic equations in rectangular form of current between node i and j
    Iij_rec = lambdify(Y_Vij, Iijrec(), 'numpy')

    :return: Iij_rec
    """
    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]
    Iij_rec = lambdify(Y_Vij, Iijrec(), 'numpy')

    return Iij_rec


def sym_func_H21_H22():
    """
    Symbolic equation of H21(∂Pi/∂θ) and H22(∂Pi/∂V)

    :return: H21i_eq_a, H21i_eq_b, H21j_eq, H22i_eq_a, H22i_eq_b, H22j_eq
    """
    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]

    H21i_eq_a = lambdify(Y_Vij, H21i_a(), 'numpy')
    H21i_eq_b = lambdify(Y_Vij, H21i_b(), 'numpy')
    H21j_eq = lambdify(Y_Vij, H21j(), 'numpy')
    H22i_eq_a = lambdify(Y_Vij, H22i_a(), 'numpy')
    H22i_eq_b = lambdify(Y_Vij, H22i_b(), 'numpy')
    H22j_eq = lambdify(Y_Vij, H22j(), 'numpy')

    return H21i_eq_a, H21i_eq_b, H21j_eq, H22i_eq_a, H22i_eq_b, H22j_eq


def sym_func_H31_H32():
    """
    Symbolic equation of H31(∂Qi/∂θ) and H32(∂Qi/∂V)

    :return: H31i_eq_a, H31i_eq_b, H31j_eq, H32i_eq_a, H32i_eq_b, H32j_eq
    """
    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]

    # Ec. de derivada de Q_i con respecto al tetha i cuando i=i y phase = phase
    H31i_eq_a = lambdify(Y_Vij, H31i_a(), 'numpy')
    H31i_eq_b = lambdify(Y_Vij, H31i_b(), 'numpy')

    # Ec. de derivada de Q_i con respecto al tetha j cuando i!=i
    # Ec(20)
    H31j_eq = lambdify(Y_Vij, H31j(), 'numpy')
    H32i_eq_a = lambdify(Y_Vij, H32i_a(), 'numpy')
    H32i_eq_b = lambdify(Y_Vij, H32i_b(), 'numpy')
    H32j_eq = lambdify(Y_Vij, H32j(), 'numpy')

    return H31i_eq_a, H31i_eq_b, H31j_eq, H32i_eq_a, H32i_eq_b, H32j_eq


def sym_func_H41_H42():
    """
    Symbolic equation of H41(∂Pij/∂θ) and H42(Pij/∂V)

    :return: H41i_eq, H41j_eq, H42i_eq, H42j_eq
    """
    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]

    H41i_eq = lambdify(Y_Vij, H41i(), 'numpy')
    H41j_eq = lambdify(Y_Vij, H41j(), 'numpy')
    H42i_eq = lambdify(Y_Vij, H42i(), 'numpy')
    H42j_eq = lambdify(Y_Vij, H42j(), 'numpy')

    return H41i_eq, H41j_eq, H42i_eq, H42j_eq


def sym_func_H51_H52():
    """
    Symbolic equation of H51(∂Qij/∂θ) and H52(Qij/∂V)

    :return: H51i_eq, H51j_eq, H52i_eq, H52j_eq
    """
    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]

    H51i_eq = lambdify(Y_Vij, H51i(), 'numpy')
    H51j_eq = lambdify(Y_Vij, H51j(), 'numpy')
    H52i_eq = lambdify(Y_Vij, H52i(), 'numpy')
    H52j_eq = lambdify(Y_Vij, H52j(), 'numpy')

    return H51i_eq, H51j_eq, H52i_eq, H52j_eq


def sym_func_H61_H62():
    """
    Symbolic equation of H61(∂|Iij|/∂θ) and H62(|Iij|/∂V)

    :return: H61i_eq, H61j_eq, H62i_eq, H62j_eq
    """
    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]

    H61i_eq = lambdify(Y_Vij, H61i(), 'numpy')
    H61j_eq = lambdify(Y_Vij, H61j(), 'numpy')
    H62i_eq = lambdify(Y_Vij, H62i(), 'numpy')
    H62j_eq = lambdify(Y_Vij, H62j(), 'numpy')

    return H61i_eq, H61j_eq, H62i_eq, H62j_eq


def sym_func_H61_H62_rec():
    """
    Symbolic equation of H61(∂Iij_rec/∂θ) and H62(Iij_rec/∂V)

    :return: H61i_rec_eq, H61j_rec_eq, H62i_rec_eq, H62j_rec_eq
    """

    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]

    H61i_rec_eq = lambdify(Y_Vij, H61i_rec(), 'numpy')
    H61j_rec_eq = lambdify(Y_Vij, H61j_rec(), 'numpy')
    H62i_rec_eq = lambdify(Y_Vij, H62i_rec(), 'numpy')
    H62j_rec_eq = lambdify(Y_Vij, H62j_rec(), 'numpy')

    return H61i_rec_eq, H61j_rec_eq, H62i_rec_eq, H62j_rec_eq


def sym_func_H01_H02():
    """
    Symbolic equation of H01(∂δij/∂θ) and H02(∂δij/∂V)

    :return: H01i_eq, H01j_eq, H02i_eq,  H02j_eq
    """

    Y_Vij = [G, B, V_i, theta_i, V_j, theta_j]

    H01i_eq = lambdify(Y_Vij, H01i(), 'numpy')
    H01j_eq = lambdify(Y_Vij, H01j(), 'numpy')
    H02i_eq = lambdify(Y_Vij, H02i(), 'numpy')
    H02j_eq = lambdify(Y_Vij, H02j(), 'numpy')

    return H01i_eq, H01j_eq, H02i_eq, H02j_eq


def func_lambdify():
    datos = [G, B, V_i, theta_i, V_j, theta_j]

    f1 = lambdify(datos, P_i(), 'numpy')
    f2 = lambdify(datos, Q_i(), 'numpy')

    f3 = lambdify(datos, P_ij(), 'numpy')
    f4 = lambdify(datos, Q_ij(), 'numpy')

    f5 = lambdify(datos, Iij(), 'numpy')
    f5_ang = lambdify(datos, angIij(), 'numpy')

    'H21(∂Pi /∂θ) and H22(∂Pi /∂V)'
    f5_a = lambdify(datos, H21i_a(), 'numpy')
    f5_b = lambdify(datos, H21i_b(), 'numpy')
    f6 = lambdify(datos, H21j(), 'numpy')
    f7_a = lambdify(datos, H22i_a(), 'numpy')
    f7_b = lambdify(datos, H22i_b(), 'numpy')
    f8 = lambdify(datos, H22j(), 'numpy')

    f9_a = lambdify(datos, H31i_a(), 'numpy')
    f9_b = lambdify(datos, H31i_b(), 'numpy')
    f10 = lambdify(datos, H31j(), 'numpy')
    f11_a = lambdify(datos, H32i_a(), 'numpy')
    f11_b = lambdify(datos, H32i_b(), 'numpy')
    f12 = lambdify(datos, H32j(), 'numpy')

    f13 = lambdify(datos, H41i(), 'numpy')
    f14 = lambdify(datos, H41j(), 'numpy')
    f15 = lambdify(datos, H42i(), 'numpy')
    f16 = lambdify(datos, H42j(), 'numpy')

    f17 = lambdify(datos, H51i(), 'numpy')
    f18 = lambdify(datos, H51j(), 'numpy')
    f19 = lambdify(datos, H52i(), 'numpy')
    f20 = lambdify(datos, H52j(), 'numpy')

    f21 = lambdify(datos, H61i(), 'numpy')
    f22 = lambdify(datos, H61j(), 'numpy')
    f23 = lambdify(datos, H62i(), 'numpy')
    f24 = lambdify(datos, H62j(), 'numpy')

    f25 = lambdify(datos, H01i(), 'numpy')
    f26 = lambdify(datos, H01j(), 'numpy')
    f27 = lambdify(datos, H02i(), 'numpy')
    f28 = lambdify(datos, H02j(), 'numpy')

    return f1, f2, f3, f4, f5, f5_ang, f5_a, f5_b, f6, f7_a, f7_b, f8, f9_a, f9_b, f10, f11_a, f11_b, f12, f13, f14, f15, f16, f17, f18, f19, f20, f21, f22, f23, f24, f25, f26, f27, f28



# -*- coding: utf-8 -*-
# @Time    : 18/08/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import logging
import os
import cmath

import pathlib
import matplotlib.pyplot as plt
import seaborn as sns
from timeit import default_timer
from functools import reduce

from .base_DSSE import *
from .MEAS_from_OpenDSS import Normalization_MEAS
from .YBus_Matrix_Pos_Seq import *
from .DSSE_algorithms.WLS_alg_1ph import WLS_1ph_state_estimator
from .DSSE_algorithms.WLS_alg_Pos_Seq import WLS_Pos_state_estimator
from .error_handling_logging import update_logg_file, to_Excel_MEAS

pd.options.mode.chained_assignment = None

#dss = py_dss_interface.DSSDLL()
data_DSS = OpenDSS_data_collection()
log_py = logging.getLogger(__name__)
orient, indent = "index", 1


def DSS_EST_save_view(
        results_DSSE: dict, MEAS_dict: dict(),
        path_save: str, name_project: str,
        DSS_coll: bool, View_res: bool, summary: bool):

    if View_res:
        print('-' * 64)
        print(f'{name_project}\nState estimation was successful\nResults obtained:')


    if DSS_coll:
        df_DSS_EST = pd.merge(results_DSSE['df_DSS'], results_DSSE['df_EST'], on=('bus_name', 'Bus Nro.'))
    else:
        df_DSS_EST = results_DSSE['df_EST']

    if path_save is not None:
        df_DSS_EST.to_excel(f'{path_save}\Results_DSSE_{name_project}.xlsx', index=False)

        df_DSS_EST_json = df_DSS_EST.to_json(orient=orient, indent=indent)
        with open(f"{path_save}\Results_DSSE_{name_project}.json", "w") as outfile:
            outfile.write(df_DSS_EST_json)

    if View_res and summary:
        print(f"Number of iterations: {results_DSSE['n_Iter']}")
        tol = format(results_DSSE['tol'], '.2E')
        print(f"Tolerance obtained :{tol}")
        print(f"Simulation time: {format(results_DSSE['time'], '.2E')} seg")
        list_aux = list()
        for x, y in results_DSSE['num_MEAS'].items():
            list_aux.append(f"|| {x}: {y}")
        print("Type and number of measurements:")
        print(*list_aux)

    if View_res == True:
        print(df_DSS_EST)

    if to_Excel_MEAS == True:
        df_MEAS_SCADA = MEAS_dict['SCADA']
        df_MEAS_PMU = MEAS_dict['PMU']
        if path_save != '':
            df_MEAS_SCADA.to_excel(f'{path_save}\MEAS_SCADA_{name_project}.xlsx', index=False)

        if (path_save != '') and (df_MEAS_PMU.empty == False):
            df_MEAS_PMU.to_excel(f'{path_save}\MEAS_PMU_{name_project}.xlsx', index=False)

def DSS_V_Ang_opt(option: list):
    df_V_Ang_dss = data_DSS.Volt_Ang_auxDSS_PU()
    df_V_Ang_dss = df_V_Ang_dss.drop(['num_nodes', 'ph_1', 'ph_2', 'ph_3'], axis=1)

    if option == alg_mapping['op1'] or \
            option == alg_mapping['op2'] or \
            option == alg_mapping['op3'] or \
            option == alg_mapping['op4'] or \
            option == alg_mapping['op5']:

        df_V_Ang_dss = df_V_Ang_dss.drop(['V2(pu)', 'V3(pu)', 'Ang2(deg)', 'Ang3(deg)'], axis=1)
        df_V_Ang_dss = df_V_Ang_dss.rename(columns={'V1(pu)': 'V1(pu)_DSS', 'Ang1(deg)': 'Ang1(deg)_DSS'})

    return df_V_Ang_dss

def EST_add_name_DSS(df_V_ang: pd.DataFrame):
    data_aux = data_DSS.allbusnames_aux().rename(columns={'bus_name_aux': 'Bus Nro.'})
    df_V_ang = data_aux.merge(df_V_ang, left_on='Bus Nro.', right_on='Bus Nro.')
    return df_V_ang


class DSSE_algorithms(BaseAlgorithm, File_options, Algorithm_options):
    """
    Class that selects the type of the state estimation algorithm
    """
    def __init__(self,
                 Sbas3ph_MVA: float, tolerance: float, max_iter: int, init_values: str,
                 DSS_path: str, MEAS_path: str, path_save: str, Typ_cir: str, ALG: str, coord: str, method: str,
                 name_project: str, MEAS_Pos: str, DSS_coll: bool, View_res: bool, summary: bool):

        BaseAlgorithm.__init__(self, Sbas3ph_MVA, tolerance, max_iter, init_values)
        File_options.__init__(self, DSS_path, MEAS_path, path_save, MEAS_Pos, name_project)
        Algorithm_options.__init__(self, Typ_cir, ALG, coord, method)
        self.DSS_coll = DSS_coll
        self.View_res = View_res
        self.summary = summary

    def addresses_algorithm(self):
        option = [self.Typ_cir, self.ALG, self.coord, self.method]

        Alg_Typ = diff_algorithms(
            self.Sbas3ph_MVA, self.tolerance, self.max_iter, self.init_values,
            self.DSS_path, self.MEAS_path, self.path_save, self.MEAS_Pos, self.name_project,
            alg_opt=option)

        if option == alg_mapping['op1'] or option == alg_mapping['op2'] or option == alg_mapping['op3']:
            state_estimation, MEAS_study = Alg_Typ.WLS_1ph_NV_Polar_linear_nonlinear_PMU(self.View_res)
        elif option == alg_mapping['op4'] or option == alg_mapping['op5']:
            state_estimation, MEAS_study = Alg_Typ.WLS_Pos_NV_Polar_nonlinear_PMU(self.View_res)
        else:
            update_logg_file('Select the configuration: Typ_circ, Alg, coord and method available. See documentation', 4, log_py)
            exit()

        DSS_EST_save_view(
            state_estimation, MEAS_study,
            self.path_save, self.name_project, self.DSS_coll, self.View_res, self.summary)

        return state_estimation


class diff_algorithms(BaseAlgorithm, File_options):

    def __init__(self,
                 Sbas3ph_MVA, tolerance, max_iter, init_values,
                 DSS_path, MEAS_path, path_save, MEAS_Pos, name_project,
                 alg_opt: list):

        BaseAlgorithm.__init__(self, Sbas3ph_MVA, tolerance, max_iter, init_values)
        File_options.__init__(self, DSS_path, MEAS_path, path_save, MEAS_Pos, name_project)
        self.alg_opt = alg_opt

    def WLS_1ph_NV_Polar_linear_nonlinear_PMU(self, View_res: bool):
        # Empty objects
        Results_DSSE = dict()
        MEAS_DSSE = dict()
        'Connecting to OpenDSS'
        dss.text(f'compile [{self.DSS_path}]')
        dss.solution_solve()

        if View_res:
            update_logg_file('State estimation of a single-phase distribution network ', 1, log_py)

        'Obtain initial values of voltage and angle'
        df_Volt_Ang_init = data_DSS.initial_voltage_angle_option(option=self.init_values)
        df_Volt_Ang_init = df_Volt_Ang_init.drop(
            ['num_nodes', 'ph_2', 'ph_3', 'V2(pu)', 'V3(pu)', 'Ang2(rad)', 'Ang3(rad)'], axis=1)

        'Get circuit YBus from OpenDSS'
        elem_YBus = ['vsources', 'lines', 'transformers']
        #initialize class
        YBus_1ph = YBus_Matrix_SeqPos_OpenDSS(data_DSS.allbusnames_aux())
        Ybus_PU = YBus_1ph.YBus_Matrix_pu(SbasMVA_3ph=self.Sbas3ph_MVA,
                                          YBus_Matrix=YBus_1ph.build_YBus_Matrix_Pos_Seq(element=elem_YBus))

        'Obtain measurements according to type and configuration'
        #initialize class
        get_MEAS = Normalization_MEAS(MEAS_path=self.MEAS_path,
                                      MEAS_Pos=self.MEAS_Pos,
                                      Sbas3ph_MVA=self.Sbas3ph_MVA,
                                      path_save=self.path_save,
                                      alg_opt=self.alg_opt)
        df_MEAS_SCADA = get_MEAS.SCADA_MEAS()
        df_MEAS_PMU = get_MEAS.PMU_MEAS()

        if self.alg_opt == alg_mapping['op1']:
            df_MEAS_PMU = pd.DataFrame(columns=df_MEAS_PMU.columns)

        # initialize class
        Alg_1ph = WLS_1ph_state_estimator(
            tolerance=self.tolerance, num_iter_max=self.max_iter, DF_MEAS_SCADA=df_MEAS_SCADA,
            DF_MEAS_PMU=df_MEAS_PMU, DF_Volt_Ang_initial=df_Volt_Ang_init, YBus_Matrix=Ybus_PU)

        if self.alg_opt == alg_mapping['op1'] or self.alg_opt == alg_mapping['op2']:
            started = default_timer()
            df_V_ang, n_Iter, tol, num_MEAS = Alg_1ph.WLS_nonlinear_linear_PMU()
            finished = default_timer()
            time = finished - started

        if self.alg_opt == alg_mapping['op3']:
            started = default_timer()
            arr_V_ang, df_V_ang, n_Iter, tol, num_MEAS = Alg_1ph.WLS_nonlinear_PMU()
            finished = default_timer()
            time = finished - started

        Results_DSSE['df_EST'] = EST_add_name_DSS(df_V_ang)
        Results_DSSE['df_DSS'] = DSS_V_Ang_opt(self.alg_opt)
        Results_DSSE['n_Iter'] = n_Iter
        Results_DSSE['tol'] = tol
        Results_DSSE['time'] = time
        Results_DSSE['num_MEAS'] = num_MEAS

        MEAS_DSSE['SCADA'] = df_MEAS_SCADA
        MEAS_DSSE['PMU'] = df_MEAS_PMU

        return Results_DSSE, MEAS_DSSE


    def WLS_Pos_NV_Polar_nonlinear_PMU(self, View_res: bool):
        # Empty objects
        Results_DSSE = dict()
        MEAS_DSSE = dict()
        'Connecting to OpenDSS'
        dss.text(f'compile [{self.DSS_path}]')
        'creation of the positive sequence circuit'
        dss.text('MakePosSeq')
        dss.text(f'Save Circuit Dir=({os.path.abspath(os.getcwd())}\PosSeq circuit)')
        SeqPos_address = f'{os.path.abspath(os.getcwd())}\PosSeq circuit\Master.DSS'
        dss.text(f'compile [{SeqPos_address}]')
        dss.solution_solve()
        '**************************************************************************************************************'
        '_______________________________________________START -> STATE ESTIMATOR_______________________________________'
        if View_res:
            update_logg_file('State estimation through the positive sequence equivalent circuit', 2, log_py)

        'Obtain initial values of voltage and angle'
        df_Volt_Ang_init = data_DSS.initial_voltage_angle_option(option=self.init_values)
        df_Volt_Ang_init = df_Volt_Ang_init.drop(
            ['num_nodes', 'ph_2', 'ph_3', 'V2(pu)', 'V3(pu)', 'Ang2(rad)', 'Ang3(rad)'], axis=1)
        'Get circuit YBus from OpenDSS'
        elem_YBus = ['vsources', 'lines', 'transformers', 'capacitors']
        YBus_SeqPos = YBus_Matrix_SeqPos_OpenDSS(data_DSS.allbusnames_aux())  # start the class
        Ybus_PU = YBus_SeqPos.YBus_Matrix_pu(
            SbasMVA_3ph=self.Sbas3ph_MVA,
            YBus_Matrix=YBus_SeqPos.build_YBus_Matrix_Pos_Seq(element=elem_YBus))

        'Obtain measurements according to type and configuration as per study_case_DSSE.py'
        get_MEAS = Normalization_MEAS(MEAS_path=self.MEAS_path,
                                      MEAS_Pos=self.MEAS_Pos,
                                      Sbas3ph_MVA=self.Sbas3ph_MVA,
                                      path_save=self.path_save,
                                      alg_opt=self.alg_opt)

        df_MEAS_SCADA, df_MEAS_PMU = get_MEAS.SCADA_PMU_MEAS_Pos()

        if self.alg_opt == ['Pos', 'NV', 'polar', 'nonlinear']:
            df_MEAS_PMU = pd.DataFrame(columns=df_MEAS_PMU.columns)

        Alg_Pos = WLS_Pos_state_estimator(
            tolerance=self.tolerance, num_iter_max=self.max_iter, DF_MEAS_SCADA=df_MEAS_SCADA,
            DF_MEAS_PMU=df_MEAS_PMU, DF_Volt_Ang_initial=df_Volt_Ang_init, YBus_Matrix=Ybus_PU)

        if self.alg_opt == alg_mapping['op4']:
            started = default_timer()
            df_V_ang, n_Iter, tol, num_MEAS = Alg_Pos.WLS_linear_nonlinear_PMU()
            finished = default_timer()
            time = finished - started

        elif self.alg_opt == alg_mapping['op5']:
            started = default_timer()
            df_V_ang, n_Iter, tol, num_MEAS = Alg_Pos.WLS_nonlinear_PMU()
            finished = default_timer()
            time = finished - started

        Results_DSSE['df_EST'] = EST_add_name_DSS(df_V_ang)
        Results_DSSE['df_DSS'] = DSS_V_Ang_opt(self.alg_opt)
        Results_DSSE['n_Iter'] = n_Iter
        Results_DSSE['tol'] = tol
        Results_DSSE['time'] = time
        Results_DSSE['num_MEAS'] = num_MEAS

        MEAS_DSSE['SCADA'] = df_MEAS_SCADA
        MEAS_DSSE['PMU'] = df_MEAS_PMU

        return Results_DSSE, MEAS_DSSE








# -*- coding: utf-8 -*-
# @Time    : 03/03/2022
# @Author  : Ing. Jorge Lara
# @Email   : jlara@iee.unsj.edu.ar
# @File    : ------------
# @Software: PyCharm

import py_dss_interface
import numpy as np
import pandas as pd
from typing import Dict, Tuple, List
from .OpenDSS_data_extraction import *


#dss = py_dss_interface.DSSDLL()

class YBus_Matrix_SeqPos_OpenDSS:

    def __init__(self, DF_allbusnames_aux: pd.DataFrame):
        self.DF_allbusnames_aux = DF_allbusnames_aux

    def build_YBus_Matrix_Pos_Seq(self, element: List[str]) -> np.array:

        nbus = dss.circuit_num_buses()
        Y_Bus = np.zeros((nbus, nbus), dtype=complex)
        DF = self.DF_allbusnames_aux

        for ii in element:
            if ii == 'vsources':
                num_element = dss.vsources_count()
                dss.vsources_first()
            elif ii == 'transformers':
                num_element = dss.transformers_count()
                dss.transformers_first()
            elif ii == 'lines':
                num_element = dss.lines_count()
                dss.lines_first()
            elif ii == 'loads':
                num_element = dss.loads_count()
                dss.loads_first()
            elif ii == 'capacitors':
                num_element = dss.capacitors_count()
                dss.capacitors_first()

            for num in range(num_element):
                if ii == 'loads':
                    bus2 = ''
                else:
                    bus2 = dss.cktelement_read_bus_names()[1]
                num_cond, node_order, bus1, bus2 = dss.cktelement_num_conductors(), dss.cktelement_node_order(), \
                                                   dss.cktelement_read_bus_names()[0], bus2
                from_bus, to_bus = from_to_bus(ii, num_cond, node_order, bus1, bus2)

                if ii == 'loads':
                    if dss.cktelement_num_phases() == 1:
                        n_Gii, n_Bii = 0, 1
                        n_Gij, n_Bij = 2, 3
                        n_Gji, n_Bji = 4, 5
                        n_Gjj, n_Bjj = 6, 7
                        Yii = np.array([[complex(dss.cktelement_y_prim()[n_Gii], (dss.cktelement_y_prim()[n_Bii]))]])
                        Yij = np.array([[complex(dss.cktelement_y_prim()[n_Gij], (dss.cktelement_y_prim()[n_Bij]))]])
                        Yji = np.array([[complex(dss.cktelement_y_prim()[n_Gji], (dss.cktelement_y_prim()[n_Bji]))]])
                        Yjj = np.array([[complex(dss.cktelement_y_prim()[n_Gjj], (dss.cktelement_y_prim()[n_Bjj]))]])
                else:
                    if dss.cktelement_num_phases() == 1:
                        n_Gii, n_Bii = 0, 1
                        n_Gij, n_Bij = 2, 3
                        n_Gji, n_Bji = 4, 5
                        n_Gjj, n_Bjj = 6, 7
                        Yii = np.array([[complex(dss.cktelement_y_prim()[n_Gii], (dss.cktelement_y_prim()[n_Bii]))]])
                        Yij = np.array([[complex(dss.cktelement_y_prim()[n_Gij], (dss.cktelement_y_prim()[n_Bij]))]])
                        Yji = np.array([[complex(dss.cktelement_y_prim()[n_Gji], (dss.cktelement_y_prim()[n_Bji]))]])
                        Yjj = np.array([[complex(dss.cktelement_y_prim()[n_Gjj], (dss.cktelement_y_prim()[n_Bjj]))]])

                if from_bus != '' and to_bus != '':
                    aux_i = (int(DF[DF['bus_name'] == from_bus]['bus_name_aux']) - 1)
                    aux_j = (int(DF[DF['bus_name'] == to_bus]['bus_name_aux']) - 1)
                elif from_bus != '' and to_bus == '':
                    aux_i = (int(DF[DF['bus_name'] == from_bus]['bus_name_aux']) - 1)
                    aux_j = ''

                if aux_i == aux_j or aux_j == '':
                    # Yii
                    Y_Bus[aux_i, aux_i] = Y_Bus[aux_i, aux_i] + Yii
                else:
                    # Yii
                    # Yii
                    Y_Bus[aux_i, aux_i] = Y_Bus[aux_i, aux_i] + Yii
                    # Yij
                    Y_Bus[aux_i, aux_j] = Y_Bus[aux_i, aux_j] + Yij
                    # Yji
                    Y_Bus[aux_j, aux_i] = Y_Bus[aux_j, aux_i] + Yji
                    # Yjj
                    Y_Bus[aux_j, aux_j] = Y_Bus[aux_j, aux_j] + Yjj

                if ii == 'transformers':
                    dss.transformers_next()
                elif ii == 'lines':
                    dss.lines_next()
                elif ii == 'loads':
                    dss.loads_next()
                elif ii == 'capacitors':
                    dss.capacitors_next()
                elif ii == 'vsources':
                    dss.vsources_next()

        return Y_Bus

    def YBus_Matrix_pu(self, SbasMVA_3ph: float, YBus_Matrix: np.array) -> np.array:
        """
        Transform YBus_Matrix into values per unit

        :param dss: COM interface between OpenDSS and Python: Electrical Network
        :param SbasMVA_3ph: Circuit Base Power
        :param YBus_Matrix: Return the function -> build_YBus_Matrix
        :param DF_allbusnames_aux: Return the function: allbusnames_aux
        :return: YBus_Matrix
        """

        SbasMVA_1ph = SbasMVA_3ph / 3
        DF = self.DF_allbusnames_aux

        for n in range(len(DF)):
            # bus_name = DF['bus_name'][n]
            dss.circuit_set_active_bus(DF['bus_name'][n])
            kVbase = dss.bus_kv_base()
            Zbas = pow(kVbase, 2) / SbasMVA_1ph
            Ybas = 1 / Zbas
            bus = (DF['bus_name_aux'][n] - 1)
            YBus_Matrix[bus, :] = YBus_Matrix[bus, :] / Ybas

        return YBus_Matrix


def get_G_B_Pos_seq(m: int, k: int, YBus_Matrix:np.array):
    """
    Function to search with coordinates m, k in YBus Matrix

    :param m: position "m" in the matrix
    :param k: position "k" in the matrix
    :param YBus_Matrix: Return of the function -> build_YBus_Matrix
    :return: G, B, G_matrix, B_matrix
    """
    Y_Matrix = YBus_Matrix[m, k]

    G = Y_Matrix.real
    B = Y_Matrix.imag

    return G, B

def Get_YMatrix_Pos_seq(DF_sequence: pd.DataFrame, DF_allbusnames_aux: pd.DataFrame):

    DF_aux = DF_allbusnames_aux
    DF_sequence.loc[:, 'from_bus_aux'] = ''
    DF_sequence.loc[:, 'to_bus_aux'] = ''
    DF_sequence.loc[:, 'B/2'] = DF_sequence['B1']
    DF_sequence['B/2'] = pd.to_numeric(DF_sequence['B/2'], errors='coerce')
    DF_sequence['B/2'] = DF_sequence['B/2'] / 2
    DF_sequence.loc[:, 'TAP'] = 1

    for index, row in DF_sequence.iterrows():
        DF_sequence.at[index, 'from_bus_aux'] = int(DF_aux[DF_aux['bus_name'] == DF_sequence['from_bus'][index]]['bus_name_aux'])
        DF_sequence.at[index, 'to_bus_aux'] = int(DF_aux[DF_aux['bus_name'] == DF_sequence['to_bus'][index]]['bus_name_aux'])

    line_data = np.array(DF_sequence)
    fb = line_data[:, 17]  # From bus number...
    tb = line_data[:, 18]  # To bus number...
    R1 = line_data[:, 11]  # Resistance, R1...
    X1 = line_data[:, 12]  # Reactance, X1...
    B1 = np.array([line_data[:, 19]]).T  # Ground Admittance, B/2...
    A1 = np.array([line_data[:, 20]]).T  # Tap setting value..
    Z1 = np.array(R1 + 1j * X1)  # Z matrix...
    Y1 = np.array([1 / Z1]).T  # To get inverse of each element...
    B1 = 1j * B1

    nbus = max(max(np.array(line_data[:, 17], dtype=int)), max(np.array(line_data[:, 18], dtype=int)))
    nbranch = len(fb)
    Ysec_bus = np.zeros((nbus, nbus), dtype=complex)

    # Formation of the Off Diagonal Elements...
    for k in range(nbranch):
        Ysec_bus[int(fb[k] - 1), int(tb[k] - 1)] = Ysec_bus[int(fb[k] - 1), int(tb[k] - 1)] - Y1[k] / A1[k]
        Ysec_bus[int(tb[k] - 1), int(fb[k] - 1)] = Ysec_bus[int(fb[k] - 1), int(tb[k] - 1)]

    # Formation of Diagonal Elements....
    for m in range(nbus):
        for n in range(nbranch):
            if fb[n] - 1 == m:
                Ysec_bus[m, m] = Ysec_bus[m, m] + (Y1[n] / A1[n] ** 2) + B1[n]

            elif tb[n] - 1 == m:
                Ysec_bus[m, m] = Ysec_bus[m, m] + Y1[n] + B1[n]

    return Ysec_bus




